/*
	SOS Emulator - [S]imple [O]S [S]cheduler Emulator
	(c) Mikhail Panasyuk aka BeAR, 2009. E-mail: otoko@yandex.ru

	Contains portions of NewOS kernel code:

	Copyright 2001-2002, Travis Geiselbrecht. All rights reserved. 
	Distributed under the terms of the NewOS License. 
	

	Original concept and results' interpretation:
	
		SOS Emulator emulates workflow of system scheduler and runs
		SOS Tester-like tests to see how this scheduler performs.
		
		See SOS Tester's descriprion (in sos_tester.cpp) for details.
		
		Schedulers are implemented as add-ons. For more info see example
		addon(s) in /add-ons folder.
	
	
	Use	at your own risk, this program may cause system crash or even burn your
	CPU(s), no warranty is provided whatsoever.
	
	Compile like this: gcc sos_emulator.cpp -lbe -lroot -o sos_emulator
	Run like this:
		sos_emulator -t 6000 -n 60 -p 30 99 -a NewOS >sos_emulator_output.txt
	or simply:
		sos_emulator
	
	Have fun! ;)
*/


#include <OS.h>
#include <String.h>
#include <List.h>
#include <image.h>
#include <Entry.h>
#include <Directory.h>
#include <Path.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <getopt.h>

#include "sos_sched.h"

/* Defaults section -- Begin */

// length of lifecycle
// milliseconds
#define DEATH_TIMER_INT32 10000 // 10 seconds
#define MIN_DEATH_TIMER_INT32 1 // 1 millisecond
#define MAX_DEATH_TIMER_INT32 3600000 // 1 hour

// threads' number regulations
#define NUM_THREADS 20
#define MIN_NUM_THREADS 2
#define MAX_NUM_THREADS 4096

// threads' priority regulations
#define L_PRIORITY 30
#define MIN_L_PRIORITY 1
#define MAX_L_PRIORITY 120

#define H_PRIORITY 99
#define MIN_H_PRIORITY 1
#define MAX_H_PRIORITY 120

#define PRIORITY_FORMULA(i,n,low,high) (int)(low*1.0+((high-low)*1.0/(n-1))*i)

// scheduler timings
// microseconds
#define SCHED_QUANTUM 3000 // 3 milliseconds
#define MIN_SCHED_QUANTUM 3
#define MAX_SCHED_QUANTUM 1000000 // 1 second

#define SCHED_PENALTY 0
#define MIN_SCHED_PENALTY 0
#define MAX_SCHED_PENALTY 1000000 // 1 second

/* Defaults section -- End */


#define IS_IN_RANGE(i,min,max) (i >= min && i <= max)

typedef struct thread sos_thread_info;

bigtime_t sched_time;
bigtime_t exec_time;

sos_thread_info idle_thread;
sos_thread_info threads[MAX_NUM_THREADS];

// Options

bigtime_t deathTimer;
int lowestPriority;
int highestPriority;
int threadsNumber;
int threadsSpawned;
bigtime_t schedQuantum;
bigtime_t schedPenalty;

bool snoozeQuantums;
bool showThreadInfo;
bool showTimeInfo;

/* SOS Emulator Addons -- Begin */

typedef struct sos_addon {
	image_id	image;
	char*		name;
} sos_addon;

BList addons;
sos_addon *selectedAddon;
	
struct thread* (*thread_get_current_thread)(void);
void (*scheduler_enqueue_in_run_queue)(struct thread *thread);
void (*scheduler_remove_from_run_queue)(struct thread *thread);
void (*scheduler_reschedule)(void);

sos_addon* find_addon(const char* name)
{
	sos_addon* currAddon;
	int count = addons.CountItems();
	for (int i = 0; i < count; i++) {
		currAddon = (sos_addon *)addons.ItemAt(i);
		if (strcmp(name, currAddon->name) == 0)
			return currAddon;
	}
	return NULL;
}

void load_addon_symbols(image_id &addon_image)
{
	get_image_symbol(addon_image, "thread_get_current_thread",
		B_SYMBOL_TYPE_TEXT, (void**)&thread_get_current_thread);
	get_image_symbol(addon_image, "scheduler_enqueue_in_run_queue",
		B_SYMBOL_TYPE_TEXT, (void**)&scheduler_enqueue_in_run_queue);
	get_image_symbol(addon_image, "scheduler_remove_from_run_queue",
		B_SYMBOL_TYPE_TEXT, (void**)&scheduler_remove_from_run_queue);
	get_image_symbol(addon_image, "scheduler_reschedule",
		B_SYMBOL_TYPE_TEXT, (void**)&scheduler_reschedule);
}

void select_addon(sos_addon *addon)
{
	load_addon_symbols(addon->image);
}

bool load_addons()
{
	image_id	addonId;
	char**		addonName;
   	status_t 	err = B_NO_ERROR; 
    BPath		path; 
	BEntry		entry;
	int32		offset;

	BDirectory	dir("add-ons");

   	//load all add-ons
	while( err == B_NO_ERROR ){
		err = dir.GetNextEntry( (BEntry*)&entry, TRUE );			
		if( entry.InitCheck() != B_NO_ERROR ){
			break;
		}
		if (entry.IsFile())
		if( entry.GetPath(&path) != B_NO_ERROR ){
			printf("entry.GetPath failed\n");
		}else{
			BString leaf(path.Leaf());
			offset = leaf.FindLast(".sos_emulator_addon");
			if (offset != B_NO_ERROR) {
				if (leaf.Length() - offset == 19) {
					addonId = load_add_on( path.Path() );
					if ( addonId < 0 ) {
						printf("load_add_on( %s ) failed\n", path.Path());
					} else {
						if( get_image_symbol( addonId, 
											"add_on_name", 
											B_SYMBOL_TYPE_DATA, 
											(void **)&addonName) ) {
							printf("get_image_symbol( add_on_name ) failed\n");
							unload_add_on( addonId );
						} else {
							sos_addon* addon = new sos_addon();
							addon->image = addonId;
							addon->name = *addonName;
							addons.AddItem((void *)addon);
						}
					}
				}
			}
		}
	}
	
	if (addons.IsEmpty()) {
		fprintf(stderr, "Error: No valid SOS Emulator add-ons were found!\n");
		return false;
	}
	return true;
}

void unload_addons()
{
	int count = addons.CountItems();
	sos_addon* addon;
	for (int i = 0; i < count; i++) {
		addon = (sos_addon *)addons.ItemAt(i);
		unload_add_on(addon->image);
		delete addon;
	}
	addons.MakeEmpty();
}

/* SOS Emulator Addons -- End */

// Print functions

void print_usage(char* appName);
void print_invalid_args(char* appName);
void print_addon_list();
void print_settings();
void print_times_table();
void print_results_table();

void get_default_settings();
bool get_user_settings(int argc, char** argv);

void sched_func(void*)
{
	struct thread *currThread;
	bigtime_t event0, event1;
	sched_time = 0;
	exec_time = 0;
	while(1)
	{
		event0 = system_time();
		
		(*scheduler_reschedule)();
		currThread = (*thread_get_current_thread)();
		
		event1 = system_time();
		
		sched_time += (event1 - event0) + schedPenalty;

		if (snoozeQuantums) {
			snooze(schedQuantum);
		}
		currThread->counter += 1;
		exec_time += schedQuantum;
	};
}

int main(int argc, char** argv)
{
	if (!load_addons()) {
		return 1;
	}
	
	get_default_settings();
	if (!get_user_settings(argc, argv)) {
		unload_addons();
		return 1;
	}
	select_addon(selectedAddon);
	print_settings();

	threadsSpawned = 0;
	for (int i=0; i<threadsNumber; i++) {
		threads[i].priority = PRIORITY_FORMULA(
			i, threadsNumber, lowestPriority, highestPriority
		);
		
		threads[i].counter = 0;
		threadsSpawned++;
	}

	set_thread_priority(find_thread(NULL), B_REAL_TIME_PRIORITY);
		// tell system who is the boss ;)

	idle_thread.priority = 0;
	idle_thread.counter = 0;
	(*scheduler_enqueue_in_run_queue)(&idle_thread);
	
	for (int i=0; i<threadsSpawned; i++) {
		(*scheduler_enqueue_in_run_queue)(&threads[i]);
	}

	thread_id sched_thread_id = spawn_thread((thread_func)sched_func,
			"scheduler emulator", 119, NULL);
	if (sched_thread_id < B_OK) {
		fprintf(stderr, "Error: can't spawn scheduler thread.");
		return 1;
	}

	resume_thread(sched_thread_id);
	snooze(deathTimer);
	kill_thread(sched_thread_id);

	if (showTimeInfo) {
		printf("\n");
		print_times_table();
	}
	if (showThreadInfo) {
		printf("\n");
		print_results_table();
	}

	unload_addons();

	return 0;
}

void print_settings()
{
	printf("Settings: work time %Lu microseconds, %i working threads.\n",
		deathTimer, threadsNumber);
	printf("Using %s scheduler add-on.\n", selectedAddon->name);
}

void print_times_table()
{
	int64 sum = sched_time + exec_time;
	printf(
		"  Task                  Time    Share (%%)\n"
		"-------------------------------------------\n"
		"  Scheduler %16Lu %12f\n"
		"  Others    %16Lu %12f\n",
		sched_time, (sched_time*1.0 / sum*1.0)*100,
		exec_time, (exec_time*1.0 / sum*1.0)*100
	);
}

void print_results_table()
{
	int64 sum = 0;
	for (int i=0; i<threadsSpawned; i++) {
		sum += threads[i].counter;
	}
	printf( "  Number  Priority            Value    Share (%%)\n"
			"--------------------------------------------------\n");
	for (int i=0; i<threadsSpawned; i++) {
		double percent = (threads[i].counter*1.0) / (sum*1.0) * 100;
		printf("%8i %9i %16Lu %12f\n",
			i, threads[i].priority, threads[i].counter, percent);
	}
}

void print_addon_list()
{
	int count = addons.CountItems();
	for (int i = 0; i < count; i++) {
		printf("%s\n", ((sos_addon *)addons.ItemAt(i))->name);
	}
}

void print_invalid_args(char* appName)
{
	fprintf(stderr, "Invalid arguments. Type '%s -h' to get help.\n", appName);
}

void print_usage(char* appName)
{
	fprintf(stderr,
		"Usage: %s [-h] [-l] [-t <time>] [-n <number>] [-p <low> <high>]"
		" [-a <name>] [-Q <time>] [-P <time>] [--no-snooze] [--no-time-info]"
		" [--no-thread-info]\n\n"
		"Defaults: -t %i -n %i -p %i %i -a %s -Q %i -P %i\n\n"
		"General:\n"
		"  -h, --help\n"
		"    : show this help\n\n"
		"  -l, --list-addons\n"
		"    : list all scheduler add-ons\n\n"
		"Test case:\n"
		"  -t, --timer <time>\n"
		"    : set work time in milliseconds,\n"
		"      <time> - between %i and %i\n\n"
		"  -n, --threads <number>\n"
		"    : set number of threads,\n"
		"      <number> - between %i and %i\n\n"
		"  -p, --priority <low> <high>\n"
		"    : set range of threads' priorities,\n"
		"      <low> - between %i and %i\n"
		"      <high> - between %i and %i\n\n"
		"Scheduler:\n"
		"  -a, --addon <name>\n"
		"    : set scheduler add-on\n\n"
		"  -Q, --quantum <time>\n"
		"    : set scheduler quantum in microseconds,\n"
		"      <time> - between %i and %i\n\n"
		"  -P, --penalty <time>\n"
		"    : set scheduler call penalty in microseconds,\n"
		"      <time> - between %i and %i\n\n"
		"  --no-snooze\n"
		"    : don't actually snooze quantums\n\n"
		"Results:\n"
		"  --no-time-info\n"
		"    : don't include time-specific info\n\n"
		"  --no-thread-info\n"
		"    : don't include thread-specific info\n\n",
		appName,
		DEATH_TIMER_INT32, NUM_THREADS, L_PRIORITY, H_PRIORITY,
		selectedAddon->name, SCHED_QUANTUM, SCHED_PENALTY,
		MIN_DEATH_TIMER_INT32, MAX_DEATH_TIMER_INT32,
		MIN_NUM_THREADS, MAX_NUM_THREADS,
		MIN_L_PRIORITY, MAX_L_PRIORITY,
		MIN_H_PRIORITY, MAX_H_PRIORITY,
		MIN_SCHED_QUANTUM, MAX_SCHED_QUANTUM,
		MIN_SCHED_PENALTY, MAX_SCHED_PENALTY
	);
}

void get_default_settings()
{
	deathTimer = DEATH_TIMER_INT32 * 1000; // conversion to microseconds
	threadsNumber = NUM_THREADS;
	lowestPriority = L_PRIORITY;
	highestPriority = H_PRIORITY;
	selectedAddon = (sos_addon *)addons.FirstItem();
	showTimeInfo = true;
	showThreadInfo = true;
	snoozeQuantums = true;
	schedQuantum = SCHED_QUANTUM;
	schedPenalty = SCHED_PENALTY;
}

bool get_user_settings(int argc, char** argv)
{
	int optc;
	
	int no_time_info = 0;
	int no_thread_info = 0;
	int no_snooze = 0;

	static struct option long_options[] = {
		{"help",			no_argument,		0,	'h'},
		{"list-addons",		no_argument,		0,	'l'},
		{"timer",			required_argument,	0,	't'},
		{"threads",			required_argument,	0,	'n'},
		{"priority",		required_argument,	0,	'p'},
		{"addon",			required_argument,	0,	'a'},
		{"quantum",			required_argument,	0,	'Q'},
		{"penalty",			required_argument,	0,	'P'},
		{"no-snooze",		no_argument,		&no_snooze,			1},
		{"no-time-info",	no_argument,		&no_time_info,		1},
		{"no-thread-info",	no_argument,		&no_thread_info,	1},
		{0, 0, 0, 0}
	};
	
	// TODO: filer wrong --* arguments
	
	while ((optc = getopt_long (argc, argv, "hl?tnpaQP", long_options,
		NULL)) != EOF)
	{
		switch (optc) {
			case 'h':
			case '?':
				print_usage(argv[0]);
				return false;
			case 'l':
				print_addon_list();
				return false;
			case 't':
				if( !optarg && optind < argc) {
					int _deathTimer = atoi(argv[optind]);
					if (IS_IN_RANGE(_deathTimer,
							MIN_DEATH_TIMER_INT32, MAX_DEATH_TIMER_INT32))
					{
						deathTimer = _deathTimer * 1000;
							// conversion to microseconds
					} else {
						print_invalid_args(argv[0]);
						return false;
					}
				} else {
					print_invalid_args(argv[0]);
					return false;
				}
				break;
			case 'n':
				if( !optarg && optind < argc) {
					int _threadsNumber = atoi(argv[optind]);
					if (IS_IN_RANGE(_threadsNumber,
							MIN_NUM_THREADS, MAX_NUM_THREADS))
					{
						threadsNumber = _threadsNumber;
					} else {
						print_invalid_args(argv[0]);
						return false;
					}
				} else {
					print_invalid_args(argv[0]);
					return false;
				}
				break;
			case 'p':
				if( !optarg && optind < argc - 1) {
					int _lowestPriority = atoi(argv[optind]);
					int _highestPriority = atoi(argv[optind + 1]);
					if (IS_IN_RANGE(_lowestPriority,
							MIN_L_PRIORITY, MAX_L_PRIORITY) &&
						IS_IN_RANGE(_highestPriority,
							MIN_H_PRIORITY, MAX_H_PRIORITY) &&
						_lowestPriority <= _highestPriority)
					{
						lowestPriority = _lowestPriority;
						highestPriority = _highestPriority;
					} else {
						print_invalid_args(argv[0]);
						return false;
					}
				} else {
					print_invalid_args(argv[0]);
					return false;
				}
				break;
			case 'a':
				if( !optarg && optind < argc) {
					sos_addon *_selectedAddon = find_addon(argv[optind]);
					if (_selectedAddon) {
						selectedAddon = _selectedAddon;
					} else {
						print_invalid_args(argv[0]);
						return false;
					}
				} else {
					print_invalid_args(argv[0]);
					return false;
				}
				break;
			case 'Q':
				if( !optarg && optind < argc) {
					int _schedQuantum = atoi(argv[optind]);
					if (IS_IN_RANGE(_schedQuantum,
							MIN_SCHED_QUANTUM, MAX_SCHED_QUANTUM))
					{
						schedQuantum = _schedQuantum;
					} else {
						print_invalid_args(argv[0]);
						return false;
					}
				} else {
					print_invalid_args(argv[0]);
					return false;
				}
				break;
			case 'P':
				if( !optarg && optind < argc) {
					int _schedPenalty = atoi(argv[optind]);
					if (IS_IN_RANGE(_schedPenalty,
							MIN_SCHED_PENALTY, MAX_SCHED_PENALTY))
					{
						schedPenalty = _schedPenalty;
					} else {
						print_invalid_args(argv[0]);
						return false;
					}
				} else {
					print_invalid_args(argv[0]);
					return false;
				}
				break;
			default:
				break;
		}
	}

	if (no_snooze) {
		snoozeQuantums = false;
	}

	if (no_time_info) {
		showTimeInfo = false;
	}
	
	if (no_thread_info) {
		showThreadInfo = false;
	}

	return true;
}
